// C/C++ headers

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

// QALE headers

#include "Widget.h"
#include "Node.h"
#include "Alignment.h"
#include "Define.h"
#include "cpp_assert.h"
#include "Component.h"
#include "ComponentSupplier.h"
#include "DefineVisitor.h"
#include "PropertyVisitor.h"
#include "AttributeVisitor.h"

using std::endl;

/*****************************************************************************
 * Definition of class: Widget
 *****************************************************************************/

Widget::Widget(
   const char *widgetName,
   const Node *width,
   const Node *height,
   const Node *x,
   const Node *y,
   const Alignment *align,
   int level
)
   :
   mParent(NULL),
   mName(strcpy(new char[strlen(widgetName) + 1], widgetName)),
   mQualifiedName(NULL),
   mWidth(width),
   mHeight(height),
   mX(x),
   mY(y),
   mAlign(align),
   mNDefines(0),
   mDefines(NULL),
   mNChildren(0),
   mChildren(NULL),
   mLevel(level),
   mComponent(NULL)
{
   // Nothing to do
}

Widget::~Widget()
{
   // Must cast to non-const before deletion

   delete [] (char *)mName;
   delete [] (char *)mQualifiedName;
   delete (Node *)mWidth;
   delete (Node *)mHeight;
   delete (Node *)mX;
   delete (Node *)mY;
   delete (Alignment *)mAlign;

   int i = 0;

   for(i = 0; i < mNDefines; i ++)
      delete mDefines[i];

   delete mDefines;

   for(i = 0; i < mNChildren; i ++)
      delete mChildren[i];

   delete mChildren;

   delete mComponent;
}

void Widget::setNDefines(int count)
{
   mNDefines = count;
   mDefines  = new Define *[count];
}

void Widget::setDefine(int index, Define *theDefine)
{
   CPP_ASSERT(index >= 0 && index < mNDefines);

   mDefines[index] = theDefine;
}

void Widget::setNChildren(int count)
{
   mNChildren = count;
   mChildren  = new Widget *[count];
}

void Widget::setChild(int index, Widget *child)
{
   CPP_ASSERT(index >= 0 && index < mNChildren);

   child->setParent(this);

   mChildren[index] = child;
}

Widget *Widget::getParent()
{
   return(mParent);
}

const char *Widget::getName()
{
   return(mName);
}

Widget *Widget::getChild(const char *name)
{
   for(int i = 0; i < mNChildren; i ++)
      if(strcmp(mChildren[i]->getName(), name) == 0)
         return(mChildren[i]);

   return(NULL);
}

const Node *Widget::getDefine(const char *name)
{
   for(int i = 0; i < mNDefines; i ++)
      if(strcmp(mDefines[i]->getName(), name) == 0)
         return(mDefines[i]->getNode());

   return(NULL);
}

void Widget::acquireComponent(
   const char *lineage,
   ComponentSupplier *supplier
)
{
   setQualifiedName(lineage);

   mComponent = supplier->getComponent(lineage);

   for(int i = 0; i < mNChildren; i ++)
   {
      const char *name = mChildren[i]->getName();
      char *buffer     =
         new char[strlen(lineage) + strlen(name) + strlen(".") + 1];

      sprintf(buffer, "%s.%s", lineage, name);

      mChildren[i]->acquireComponent(buffer, supplier);

      delete [] buffer;
   }
}

Component *Widget::getComponent()
{
   return(mComponent);
}

void Widget::setQualifiedName(const char *qualifiedName)
{
   mQualifiedName = strcpy(
      new char[strlen(qualifiedName) + 1], qualifiedName
   );
}

const char *Widget::getQualifiedName() const
{
   return(mQualifiedName);
}

void Widget::handleVisitor(Visitor &v)
{
   v.setCurrentWidget(this);

   v.setField(Visitor::SIZE_WIDTH);
   mWidth->accept(v);

   v.setField(Visitor::SIZE_HEIGHT);
   mHeight->accept(v);

   v.setField(Visitor::X_LOCATION);
   mX->accept(v);

   v.setField(Visitor::Y_LOCATION);
   mY->accept(v);
}

void Widget::resolveDefines(DefineVisitor &v)
{
   handleVisitor(v);

   for(int i = 0; i < mNChildren; i ++)
      mChildren[i]->resolveDefines(v);
}

void Widget::resolveProperties(PropertyVisitor &v)
{
   handleVisitor(v);

   for(int i = 0; i < mNChildren; i ++)
      mChildren[i]->resolveProperties(v);
}

void Widget::resolveAttributes(AttributeVisitor &v)
{
   handleVisitor(v);

   for(int i = 0; i < mNChildren; i ++)
      mChildren[i]->resolveAttributes(v);
}

void Widget::setSize()
{
   double width  = mWidth->evaluate();
   double height = mHeight->evaluate();

   if(width > 0.0 && height > 0.0)
   {
      mComponent->setSize(
         width, height
      );
   }

   for(int i = 0; i < mNChildren; i ++)
      mChildren[i]->setSize();
}

void Widget::setLocation()
{
   double x = mX->evaluate();
   double y = mY->evaluate();

   // Adjust the values based upon alignment

   Alignment::Type t = mAlign->get();

   if(t == Alignment::_LT || t == Alignment::_LC || t == Alignment::_LB)
   {
      // No need to adjust x
   }
   else
   {
      double width = mComponent->getWidth();

      if(t == Alignment::_CT || t == Alignment::_CC || t == Alignment::_CB)
         x -= width / 2;

      else
         x -= width;
   }

   if(t == Alignment::_LT || t == Alignment::_CT || t == Alignment::_RT)
   {
      // No need to adjust y
   }
   else
   {
      double height = mComponent->getHeight();

      if(t == Alignment::_LC || t == Alignment::_CC || t == Alignment::_RC)
         y -= height / 2;

      else
         y -= height;
   }

   mComponent->setLocation(
      x, y
   );

   for(int i = 0; i < mNChildren; i ++)
      mChildren[i]->setLocation();
}

ostream &operator<<(ostream &out, const Widget &w)
{
   char *indent = new char[w.mLevel + 1];

   int i = 0;

   for(i = 0; i < w.mLevel; i ++)
      indent[i] = ' ';

   indent[w.mLevel] = '\0';

   out << indent << "widgetName: \"" << w.mName      << "\"" << endl;
   out << indent << "width     : \"" << *w.mWidth    << "\"" << endl;
   out << indent << "height    : \"" << *w.mHeight   << "\"" << endl;
   out << indent << "x         : \"" << *w.mX        << "\"" << endl;
   out << indent << "y         : \"" << *w.mY        << "\"" << endl;
   out << indent << "alignment : \"" << *w.mAlign    << "\"";

   if(w.mNDefines > 0)
   {
      out << endl << indent << "defines {";

      for(i = 0; i < w.mNDefines; i ++)
         out << endl << *w.mDefines[i];

      out << endl << indent << "}";
   }

   if(w.mNChildren > 0)
   {
      out << endl << indent << "children {";

      for(i = 0; i < w.mNChildren; i ++)
         out << endl << *w.mChildren[i];

      out << endl << indent << "}";
   }

   delete [] indent;

   return(out);
}

void Widget::setParent(Widget *theParent)
{
   mParent = theParent;
}
